"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.Wormhole = void 0;
const sdk_base_1 = require("@wormhole-foundation/sdk-base");
const sdk_definitions_1 = require("@wormhole-foundation/sdk-definitions");
const config_1 = require("./config");
const tokenTransfer_1 = require("./protocols/tokenTransfer");
const cctpTransfer_1 = require("./protocols/cctpTransfer");
const tasks_1 = require("./tasks");
const whscan_api_1 = require("./whscan-api");
const circle_api_1 = require("./circle-api");
class Wormhole {
    _platforms;
    _chains;
    _network;
    conf;
    constructor(network, platforms, conf) {
        this._network = network;
        this.conf = conf ?? config_1.CONFIG[network];
        this._chains = new Map();
        this._platforms = new Map();
        for (const p of platforms) {
            const platformName = p.platform;
            const filteredChains = (0, config_1.networkPlatformConfigs)(network, platformName);
            this._platforms.set(platformName, p.setConfig(network, filteredChains));
        }
    }
    get network() {
        return this._network;
    }
    /**
     * Creates a CCTPTransfer object to move Native USDC from one chain to another
     * @param amount the amount to transfer
     * @param from the address to transfer from
     * @param to the address to transfer to
     * @param automatic whether to use automatic delivery
     * @param payload the payload to send with the transfer
     * @param nativeGas the amount of native gas to send with the transfer
     * @returns the CCTPTransfer object
     * @throws Errors if the chain or protocol is not supported
     */
    async cctpTransfer(amount, from, to, automatic, payload, nativeGas) {
        if (automatic && payload)
            throw new Error("Payload with automatic delivery is not supported");
        if (!(0, sdk_base_1.isCircleChain)(from.chain) ||
            !(0, sdk_base_1.isCircleChain)(to.chain) ||
            !(0, sdk_base_1.isCircleSupported)(this.network, from.chain) ||
            !(0, sdk_base_1.isCircleSupported)(this.network, to.chain))
            throw new Error(`Network and chain not supported: ${this.network} ${from.chain} `);
        return await cctpTransfer_1.CCTPTransfer.from(this, {
            amount,
            from,
            to,
            automatic,
            payload,
            nativeGas,
        });
    }
    /**
     * Creates a TokenTransfer object to move a token from one chain to another
     * @param token the token to transfer
     * @param amount the amount to transfer
     * @param from the address to transfer from
     * @param to the address to transfer to
     * @param automatic whether to use automatic delivery
     * @param payload the payload to send with the transfer
     * @param nativeGas the amount of native gas to send with the transfer
     * @returns the TokenTransfer object
     * @throws Errors if the chain or protocol is not supported
     */
    async tokenTransfer(token, amount, from, to, automatic, payload, nativeGas) {
        // TODO: check if `toChain` is gateway supported
        // not enough to check if its a Cosmos chain since Terra/Xpla/Sei are not supported
        // if (chainToPlatform(to.chain) === 'Cosmwasm' ) {
        //   return await GatewayTransfer.from(this, {
        //     token,
        //     amount,
        //     from,
        //     to,
        //     payload,
        //     nativeGas,
        //   });
        // }
        return await tokenTransfer_1.TokenTransfer.from(this, {
            token,
            amount,
            from,
            to,
            automatic,
            payload,
            nativeGas,
        });
    }
    /**
     * Gets the contract addresses for a given chain
     * @param chain the chain name or chain id
     * @returns the contract addresses
     */
    getContracts(chain) {
        return this.conf.chains[chain]?.contracts;
    }
    /**
     * Returns the platform object, i.e. the class with platform-specific logic and methods
     * @param chain the chain name or platform name
     * @returns the platform context class
     * @throws Errors if platform is not found
     */
    getPlatform(chain) {
        const platformName = (0, sdk_base_1.isChain)(chain)
            ? this.conf.chains[chain].platform
            : chain;
        const platform = this._platforms.get(platformName);
        if (!platform)
            throw new Error(`Not able to retrieve platform ${platform}`);
        return platform;
    }
    /**
     * Returns the chain "context", i.e. the class with chain-specific logic and methods
     * @param chain the chain name
     * @returns the chain context class
     * @throws Errors if context is not found
     */
    getChain(chain) {
        if (!this._chains.has(chain))
            this._chains.set(chain, this.getPlatform(chain).getChain(chain));
        return this._chains.get(chain);
    }
    /**
     * Gets the TokenId for a token representation on any chain
     *  These are the Wormhole wrapped token addresses, not necessarily
     *  the cannonical version of that token
     *
     * @param tokenId The Token ID (chain/address)
     * @param chain The chain name or id
     * @returns The TokenId on the given chain, null if it does not exist
     * @throws Errors if the chain is not supported or the token does not exist
     */
    async getWrappedAsset(chain, token) {
        const ctx = this.getChain(chain);
        const tb = await ctx.getTokenBridge();
        return { chain, address: await tb.getWrappedAsset(token) };
    }
    /**
     * Gets the number of decimals for a token on a given chain
     *
     * @param tokenId The Token ID (home chain/address)
     * @param chain The chain name or id of the token/representation
     * @returns The number of decimals
     */
    async getDecimals(chain, token) {
        const ctx = this.getChain(chain);
        return await ctx.getDecimals(token);
    }
    /**
     * Converts a human friendly decimal number to base units for the token passed
     *
     * @param chain The chain name
     * @param tokenId The token ID (its home chain and address on the home chain) or 'native'
     * @param amount The decimal number as a string to convert into base units
     * @returns The amount converted to base units as a BigNumber
     */
    async normalizeAmount(chain, token, amount) {
        const ctx = this.getChain(chain);
        let decimals = await ctx.getDecimals(token);
        return (0, sdk_base_1.normalizeAmount)(amount, decimals);
    }
    /**
     * Fetches the balance of a given token for a wallet
     *
     * @param walletAddress The wallet address
     * @param tokenId The token ID (its home chain and address on the home chain)
     * @param chain The chain name or id
     * @returns The token balance of the wormhole asset as a BigNumber
     */
    async getBalance(chain, token, walletAddress) {
        const ctx = this.getChain(chain);
        return ctx.getBalance(walletAddress, token);
    }
    /**
     * Gets the associated token account for chains that require it (only Solana currently).
     * @param sendingChain the chain name of the source token
     * @param sendingToken the TokenId or address of the source token
     * @param recipient the address of the recipient
     * @returns
     */
    async getTokenAccount(sendingChain, sendingToken, recipient) {
        // TODO: same as supportsSendWithRelay, need some
        // way to id this in a less sketchy way
        const chain = this.getChain(recipient.chain);
        if (!("getTokenAccount" in chain))
            return recipient;
        let t;
        if (sendingToken === "native") {
            const srcTb = await this.getChain(sendingChain).getTokenBridge();
            t = { chain: sendingChain, address: await srcTb.getWrappedNative() };
        }
        else {
            t = (0, sdk_definitions_1.isTokenId)(sendingToken)
                ? sendingToken
                : {
                    chain: sendingChain,
                    address: sendingToken.toUniversalAddress(),
                };
        }
        const dstTokenBridge = await chain.getTokenBridge();
        const dstNative = await dstTokenBridge.getWrappedAsset(t);
        return {
            chain: recipient.chain,
            // @ts-ignore
            address: await chain.getTokenAccount(dstNative, recipient.address),
        };
    }
    /**
     * Gets the Raw VAA Bytes from the API or Guardian RPC, finality must be met before the VAA will be available.
     * @param chain The chain name
     * @param emitter The emitter address
     * @param sequence The sequence number
     * @param retries The number of times to retry
     * @param opts The options for the request. Timeouts must be set in ms.
     * @returns The VAA bytes if available
     * @throws Errors if the VAA is not available after the retries
     */
    async getVAABytes(chain, emitter, sequence, timeout = config_1.DEFAULT_TASK_TIMEOUT) {
        const task = () => (0, whscan_api_1.getVaaBytes)(this.conf.api, {
            chain,
            sequence,
            emitter: emitter.toUniversalAddress(),
        });
        const vaaBytes = await (0, tasks_1.retry)(task, config_1.WHSCAN_RETRY_INTERVAL, timeout, "Wormholescan:GetVaaBytes");
        return vaaBytes ?? undefined;
    }
    /**
     * Gets a VAA from the API or Guardian RPC, finality must be met before the VAA will be available.
     * @param chain The chain name
     * @param emitter The emitter address
     * @param sequence The sequence number
     * @param timeout The total amount of time to wait for the VAA to be available
     * @returns The VAA if available
     * @throws Errors if the VAA is not available after the retries
     */
    async getVAA(chain, emitter, sequence, decodeAs, timeout = config_1.DEFAULT_TASK_TIMEOUT) {
        const vaaBytes = await this.getVAABytes(chain, emitter, sequence, timeout);
        if (vaaBytes === undefined)
            return;
        return (0, sdk_definitions_1.deserialize)(decodeAs, vaaBytes);
    }
    async getCircleAttestation(msgHash, timeout = config_1.DEFAULT_TASK_TIMEOUT) {
        const task = () => (0, circle_api_1.getCircleAttestation)(this.conf.circleAPI, msgHash);
        return (0, tasks_1.retry)(task, config_1.CIRCLE_RETRY_INTERVAL, timeout, "Circle:GetAttestation");
    }
    /**
     * Get the status of a transaction, identified by the chain, emitter address, and sequence number
     *
     * @param chain the chain name
     * @param emitter the emitter address
     * @param sequence the sequence number
     * @returns the TransactionStatus
     */
    async getTransactionStatus(chain, emitter, sequence, timeout = config_1.DEFAULT_TASK_TIMEOUT) {
        const whm = {
            chain,
            emitter,
            sequence,
        };
        const task = () => (0, whscan_api_1.getTransactionStatus)(this.conf.api, whm);
        const status = await (0, tasks_1.retry)(task, config_1.WHSCAN_RETRY_INTERVAL, timeout, "Wormholescan:TransactionStatus");
        if (!status)
            throw new Error(`No data available for : ${whm}`);
        return status;
    }
    /**
     * Parse an address to a universal address
     *
     * @param address The native address
     * @returns The address in the universal format
     */
    parseAddress(chain, address) {
        return (0, sdk_definitions_1.toNative)(chain, address);
    }
    /**
     * Parses all relevant information from a transaction given the sending tx hash and sending chain
     *
     * @param tx The sending transaction hash
     * @param chain The sending chain name or id
     * @returns The parsed WormholeMessageId
     */
    async parseMessageFromTx(chain, txid, timeout) {
        const originChain = this.getChain(chain);
        const task = async () => {
            const msgs = await originChain.parseTransaction(txid);
            // possible the node we hit does not have this data yet
            // return null to signal retry
            if (msgs.length === 0)
                return null;
            return msgs;
        };
        const parsed = await (0, tasks_1.retry)(task, originChain.config.blockTime, timeout, "WormholeCore:ParseMessageFromTransaction");
        if (!parsed)
            throw new Error(`No WormholeMessageId found for ${txid}`);
        if (parsed.length != 1)
            throw new Error(`Expected a single VAA, got ${parsed.length}`);
        return parsed;
    }
}
exports.Wormhole = Wormhole;
//# sourceMappingURL=wormhole.js.map